# @package      hubzero-python
# @file         user.py
# @author       David Benham <dbenham@purdue.edu>
# @copyright    Copyright (c) 2012-2013 HUBzero Foundation, LLC.
# @license      http://www.gnu.org/licenses/lgpl-3.0.html LGPLv3
#
# Copyright (c) 2012-2013 HUBzero Foundation, LLC.
#
# This file is part of: The HUBzero(R) Platform for Scientific Collaboration
#
# The HUBzero(R) Platform for Scientific Collaboration (HUBzero) is free
# software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# HUBzero is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# HUBzero is a registered trademark of HUBzero Foundation, LLC.
#

import datetime
import hubzero.config.passwords
import hubzero.utilities.group
import hubzero.utilities.misc
import hubzero.config.webconfig
import hubzero.data.db
import ldap
import os
import sys


def userExists(username, hubname = ""):
	dbPW = hubzero.config.webconfig.getWebConfigDBPassword(hubname)
	dbUserName = hubzero.config.webconfig.getWebConfigDBUsername(hubname)
	dbName = hubzero.config.webconfig.getWebConfigDBName(hubname)

	db =  hubzero.data.db.MySQLConnection("localhost", dbName, dbUserName, dbPW)
	if _userExists(db, username):
		return True
	else:
		return False


def _userExists(db, username):
	sql = 'select id from `jos_users` where username = %s'
	data = (username)
	userID = db.query_selectscalar(sql, data)

	return userID


def userLookup(username, hubname = ""):
	dbPW = hubzero.config.webconfig.getWebConfigDBPassword(hubname)
	dbUserName = hubzero.config.webconfig.getWebConfigDBUsername(hubname)
	dbName = hubzero.config.webconfig.getWebConfigDBName(hubname)

	db =  hubzero.data.db.MySQLConnection("localhost", dbName, dbUserName, dbPW)
	rv = _userExists(db, username)
	return rv


def _getUserID(db, username):
	sql = 'select id from `jos_users` where username = %s'
	data = (username)


	userID = db.query_selectscalar(sql, data)

	if userID is None:
		return None
	else:
		return userID


def _del_jos_users(db, username):
	sql = 'select id from `jos_users` where username = %s'
	data = (username)
	userID = db.query_selectscalar(sql, data)	

	sql = 'delete from `jos_users` where userid = %s'
	data = (userID)
	userID = db.query_selectscalar(sql, data)	

	return userID


def _del_jos_core_acl_aro(db, userID):

	sql = 'select id from `jos_core_acl_aro` where `value` = %s'
	data = (userID)
	aroID = db.query_selectscalar(sql, data)	

	sql = "DELETE FROM `jos_core_acl_aro` where `value` = %s"
	data = (userID)
	db.query_rowcount(sql, data)

	return aroID


def _del_jos_core_acl_groups_aro_map(db, aroID):
	print "aroID = " + str(aroID)
	sql = "DELETE FROM `jos_core_acl_groups_aro_map` WHERE `aro_id` = %s"
	data = (aroID)
	db.query_rowcount(sql, data)


def _del_jos_xprofiles(db, uidNumber):
	sql = "DELETE FROM `jos_xprofiles` WHERE `uidNumber` = %s"
	data = (uidNumber)
	db.query_rowcount(sql, data)


def _update_jos_xprofiles_pw(db, userid, password):
	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	sql = "UPDATE `jos_xprofiles` SET `userPassword`=%s WHERE `uidNumber`=%s"
	data = (pw, userid)
	db.query_rowcount(sql, data)


def _update_jos_users_pw(db, userid, password):
	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	sql = "UPDATE `jos_users` SET `password` = %s WHERE `id`=%s"
	data = (pw, userid)
	db.query_rowcount(sql, data)


def _update_jos_users_password(db, userid, password):
	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	sql = "UPDATE `jos_users_password` SET `passhash` = %s WHERE `user_id`=%s"
	data = (pw, userid)
	db.query_rowcount(sql, data)


#TODO jos_password_history??



def _update_ldap_user_pw(username, uidNumber, password):
	"""
	Use raw pw here, we'll hash it
	"""

	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn")
	hubLDAPAcctMgrPW = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw")
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn")	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)

	dn = "uid=" + username + ",ou=users," + hubLDAPBaseDN

	attrs = [ (ldap.MOD_REPLACE, 'userPassword', pw) ]

	l.modify_s(dn, attrs)	


def _insert_jos_users(db, joomlauserid, name, username, email, password, jgidNumber):

	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to '*'? we think not

	# xlate numeric jgidNumber to the string (Joomla needs both for some reason)
	# 18=Registered, 23=Manager, 24=Administrator, 25=Super Administrator
	userTypeLookupDict = {18:"Registered", 23:"Manager", 24:"Administrator", 25:"Super Administrator"}

	if joomlauserid == -1:
		sql = "INSERT INTO `jos_users` (`name`, `username`, `email`, `password`, `usertype`, `block`, `sendEmail`, `gid`, `registerDate`, `lastvisitDate`, `activation`, `params`)"
		sql += "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);"
		data = (name, username, email, pw, userTypeLookupDict[int(jgidNumber)], 0, 1, jgidNumber, datetime.datetime.now(), '0000-00-00 00:00:00', '', '');
		userID = db.query_lastrowid(sql, data)
	else:
		sql = "INSERT INTO `jos_users` (`id`, `name`, `username`, `email`, `password`, `usertype`, `block`, `sendEmail`, `gid`, `registerDate`, `lastvisitDate`, `activation`, `params`)"
		sql += "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);"
		data = (joomlauserid, name, username, email, pw, userTypeLookupDict[int(jgidNumber)], 0, 1, jgidNumber , datetime.datetime.now(), '0000-00-00 00:00:00', '', '');
		userID = db.query_lastrowid(sql, data)

	return userID


def _insert_jos_core_acl_aro(db, userID, fullname):
	sql = "INSERT INTO `jos_core_acl_aro` (`section_value`, `value`, `order_value`, `name`, `hidden`) "
	sql += "VALUES (%s, %s, %s, %s, %s)"
	data = ( 'users',  userID, 0, fullname, 0)
	aroID = db.query_lastrowid(sql, data)
	return aroID


def _insert_jos_core_acl_groups_aro_map(db, jgidNumber, aroID):
	sql = "INSERT INTO `jos_core_acl_groups_aro_map` (`group_id`, `section_value`, `aro_id`) "
	sql += "VALUES (%s, %s, %s)"
	data = ( jgidNumber, '', aroID)
	db.query_rowcount(sql, data)


def _insert_jos_xprofiles(db, hubname, uidNumber, fname, mname, lname, username, email, password, gid, gidNumber, loginShell, homeDir):

	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	sql = "INSERT INTO `jos_xprofiles` (`uidNumber`, `name`, `username`, `email`, `registerDate`, `gidNumber`, `homeDirectory`, `loginShell`, "
	sql += "`ftpShell`, `userPassword`, `gid`, `orgtype`, `organization`, `countryresident`, `countryorigin`, `gender`, `url`, `reason`, "
	sql += "`mailPreferenceOption`, `usageAgreement`, `jobsAllowed`, `modifiedDate`, `emailConfirmed`, `regIP`, `regHost`, `nativeTribe`, `phone`, "
	sql += "`proxyPassword`, `proxyUidNumber`, `givenName`, `middleName`, `surname`, `picture`, `vip`, `public`, `params`, `note`, `shadowExpire`) "
	sql += "VALUES (" + '%s,'*37 + "%s );" 

	data = (uidNumber, fname + ' ' + lname, username, email, datetime.datetime.now(), gidNumber, homeDir, loginShell, 
		    '/usr/lib/sftp-server', pw, gid, '', '', '', '', '', '', '', 
		    0, 0, 3, '0000-00-00 00:00:00', 1, '', '', '', '',
		    '', '', fname, mname, lname, '', 0, 0, '', '', None)
	db.query_rowcount(sql, data)


def _insert_jos_users_password(db, userid, password):
	sql = "INSERT INTO jos_users_password (`user_id`, `passhash`) " 
	sql += "VALUES (%s, %s)"

	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	data = (userid, pw)
	db.query_rowcount(sql, data)


def add_ldap_user(uid, 
                  gecos, 
                  password, 
                  uidNumber, 
                  homeDir, 
                  loginShell, 
                  gid, 
                  gidNumber, 
                  hubname = ""):

	# get ldapManagerUserDN and PW

	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn", hubname)
	hubLDAPAcctMgrPW= hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw", hubname)
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn", hubname)	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)

	dn = "uid=" + str(uid) + ",ou=users," + hubLDAPBaseDN
	attrs = {}
	attrs['objectclass'] = ['top', 'account', 'posixAccount', 'shadowAccount']  
	attrs['cn'] = str(uid)

	if password != '*':
		attrs['userPassword'] = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		attrs['userPassword'] = password
	else:
		attrs['userPassword'] = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	attrs['uid'] = uid
	attrs['uidNumber'] = str(uidNumber)
	attrs['homeDirectory'] = homeDir
	#attrs['cn'] = gecos
	attrs['loginShell'] = loginShell

	# TODO: grab this from the com_members component parm field
	attrs['gidNumber'] = str(gidNumber)
	#attrs['gid'] = gid

	ldif = ldap.modlist.addModlist(attrs)
	l.add_s(dn, ldif)


def delete_ldap_user(username, hubname = ""):
	# get ldapManagerUserDN and PW
	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn", hubname)
	hubLDAPAcctMgrPW= hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw", hubname)
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn", hubname)	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)
	dn = "uid=" + username + ",ou=users," + hubLDAPBaseDN
	l.delete_s(dn)


def addhubuser(hubname, 
               username, 
               email, 
               pw,
               pwdisabled = False, 
               grouptype = 1,
               joomlagidNumber = 18, 
               gid = "users", 
               gidNumber = 100, 
               uidNumber = -1, 
               gecos = "",
               loginShell = "", 
               skipCreateHome = False, 
               homeDir = "", 
               skipCreateGroup = False, 
               skipCreateLDAPUser = False, 
               skipCreateCMSUser = False, 
               uid_min = 1000, 
               uid_max = 65533):
	"""
	hubname - for machines that have more than one hub, otherwise this will use the default hub for a server
	username - username to add
	email - user email
	pw - unencrypted password, we'll hash it as necessary using MD5
	pwdisabled - should the PW be disabled?
	joomlagidNumber - 18=Registered, 23=Manager, 24=Administrator, 25=Super Administrator
	gid - linux gid membership
	gidNumber - linux gidNumber, corresponds to the group with the previous gid
	uidNumber - specify the jos_users PK ID, also is the uidNumber for the ldap (same thing)
	gecos - comma delimited info list on user, we curretnly only use first entry, which is a full name (fname, mname, lname)
	loginShell - user's login shell
	skipCreateHome - skip creating user's home directory
	homeDir - path to user's home directory
	skipCreateGroup - skip creating a default group for new user, group name matches the username
	skipCreateLDAPUser - skip creating ldap info for this user
	skipCreateCMSUser - skip ldap info for this user
	"""

	# grab config options
	dbPW = hubzero.config.webconfig.getWebConfigDBPassword(hubname)
	dbUserName = hubzero.config.webconfig.getWebConfigDBUsername(hubname)
	dbName = hubzero.config.webconfig.getWebConfigDBName(hubname)

	db =  hubzero.data.db.MySQLConnection("localhost", dbName, dbUserName, dbPW)

	# Parse gecos, it's a massive mess, but it's a holdover from the wild west days of unix and it's a comma delimited field 
	# used in the /etc/passwd file to store personal info. It's comma delimited, and name is the first item in the list... 
	# right now we don't care about what else gets passed, just the first entry
	gecosFields = gecos.split(",")
	fullName = gecosFields[0]
	fullNameList = gecosFields[0].split(" ")

	if len(fullNameList) == 3:
		firstName = fullNameList[0]
		middleName = fullNameList[1]
		lastName = fullNameList[2]
	elif len(fullNameList) == 2:
		firstName = fullNameList[0]
		middleName = ""
		lastName = fullNameList[1]
	elif len(fullNameList) == 1:
		firstName = fullNameList[0]
		middleName = ""
		lastName = ""
	else:
		firstName = ""
		middleName = ""
		lastName = ""

	# usually homeDir will be blank, but not always
	if homeDir == "":
		homeDir = '/home/' + hubname + "/" + username

	if pwdisabled: 
		pw = "*"

	# check for unique username
	if not skipCreateCMSUser:
		# when inserting a cms user, with or without a ldap user, test the mysql database for unique usernames
		if _userExists(db, username):
			print "User '" + username + "' already exists in CMS"
			return (1)
	elif not skipCreateLDAPUser:
		# no cms and ldap, check ldap
		if uidExists(username):
			print "User '" + username + "' already exists in ldap"
			return (2)
	else:
		print "skip cms insert and ldap, but then why bother?"
		return(3)

	# if the user doesn't want custom group for this user, they still need to be added to some group
	if not skipCreateGroup:
		gidNumber = 100 # default to the users group
		gid = 'users'

	if not skipCreateCMSUser:

		# create group for user named same as username (need group insert here for xprofiles addition below)
		# Guess if user specifies a specific already created group for the new user and says create a new group for the 
		# user, we'll override and create the new group with this logic
		if not skipCreateGroup:
			gidNumber = hubzero.utilities.group._insert_jos_xgroups(db, username, username + " group", grouptype)	
			gid = username # group name same as user

		uidNumber = _insert_jos_users(db, uidNumber, fullName, username, email, pw, joomlagidNumber)
		aroID = _insert_jos_core_acl_aro(db, uidNumber, fullName)
		_insert_jos_core_acl_groups_aro_map(db, joomlagidNumber, aroID)
		_insert_jos_xprofiles(db, hubname, uidNumber, firstName, middleName, lastName, username, email, pw, gid, gidNumber, loginShell, homeDir)
		_insert_jos_users_password(db, uidNumber, pw)

		# add user to new group (need xgroups insert after user creation because we need userid, I'd prefer to keep all group logic in one place)
		if not skipCreateGroup:
			hubzero.utilities.group._insert_jos_xgroups_members(db, gidNumber, uidNumber)

		# create home directory
		if not skipCreateHome:
			rc, procStdOut, procStdErr = hubzero.utilities.misc.exShellCommand(["mkdir", homeDir, "--mode", "0750"])
			print procStdOut
			if rc:
				print procStdErr

			print "creating homedir: " + homeDir

			# set ownership
			os.chown(homeDir, uidNumber, gidNumber)

			# change permissions user full, noone else has anything
			rc, procStdOut, procStdErr = hubzero.utilities.misc.exShellCommand(["chmod", '700', homeDir])
			if rc:
				print procStdErr

	if not skipCreateLDAPUser:

		# if no uidNumber specified, find an available one
		if uidNumber == -1:
			uidNumber = hubzero.utilities.user.getLDAPMinUIDNumber(hubname, uid_min, uid_max)

		if not skipCreateGroup:
			hubzero.utilities.group.add_ldap_group(username, username + " group", gidNumber)

		add_ldap_user(username, gecos, pw, uidNumber, homeDir, loginShell, gid, gidNumber)

	return(0)


def delhubuser(username, hubname = ""):
	dbPW = hubzero.config.webconfig.getWebConfigDBPassword(hubname)
	dbUserName = hubzero.config.webconfig.getWebConfigDBUsername(hubname)
	dbName = hubzero.config.webconfig.getWebConfigDBName(hubname)
	db =  hubzero.data.db.MySQLConnection("localhost", dbName, dbUserName, dbPW)

	sql = 'select id from `jos_users` where username = %s'
	data = (username)
	userID = db.query_selectscalar(sql, data)

	print "deleting user (" + username + ") from database"

	aroID = _del_jos_core_acl_aro(db, userID)
	_del_jos_core_acl_groups_aro_map(db, aroID)
	_del_jos_xprofiles(db, userID)
	_del_jos_users(db, username)

	delete_ldap_user(username)


def updateUserPW(username, password, updateCMS = True, updateLDAP = True, hubname = ""):
	"""
	pass plaintext pw, we'll hash it
	"""

	dbPW = hubzero.config.webconfig.getWebConfigDBPassword(hubname)
	dbUserName = hubzero.config.webconfig.getWebConfigDBUsername(hubname)
	dbName = hubzero.config.webconfig.getWebConfigDBName(hubname)
	db =  hubzero.data.db.MySQLConnection("localhost", dbName, dbUserName, dbPW)

	userid = _getUserID(db, username)

	if password != '*':
		pw = hubzero.config.passwords.MD5PasswordHash(password)
	elif '{MD5}' in password:
		pw = password
	else:
		pw = '{MD5}*' # can user generate a value that will MD5 hash to, we think not '*'?

	if updateCMS:
		_update_jos_xprofiles_pw(db, userid, pw)
		_update_jos_users_pw(db, userid, pw)
		_update_jos_users_password(db, userid, pw)

	if updateLDAP:
		_update_ldap_user_pw(username, userid, pw)


def uidNumberExists(uidNumber, hubname = ""):
	"""
	See if specified uidNumber is in the ldap user ou
	"""

	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn", hubname)
	hubLDAPAcctMgrPW = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw", hubname)
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn", hubname)	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)

	results = l.search_s('ou=users,' + hubLDAPBaseDN, ldap.SCOPE_SUBTREE, '(uidNumber=' + uidNumber + ')', None)

	if not results:
		return False
	else:
		return True


def uidExists(uid, hubname = ""):
	"""
	See if specified uid is in the ldap user ou
	"""

	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn", hubname)
	hubLDAPAcctMgrPW = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw", hubname)
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn", hubname)	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)

	results = l.search_s('ou=users,' + hubLDAPBaseDN, ldap.SCOPE_SUBTREE, '(uid=' + uid + ')', None)

	if not results:
		return False
	else:
		return True


def getLDAPMinUIDNumber(hubname = "", minUIDNumber=1, maxUIDNumber=1000):
	"""
	get max(uid) of all objects of class posixAccount in ou=Users
	"""

	hubLDAPAcctMgrDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerdn", hubname)
	hubLDAPAcctMgrPW = hubzero.config.webconfig.getComponentParam("com_system", "ldap_managerpw", hubname)
	hubLDAPBaseDN = hubzero.config.webconfig.getComponentParam("com_system", "ldap_basedn", hubname)	

	l = ldap.initialize('ldap://localhost')
	l.simple_bind_s(hubLDAPAcctMgrDN, hubLDAPAcctMgrPW)

	results = l.search_s('ou=users,' + hubLDAPBaseDN, ldap.SCOPE_SUBTREE, '(objectclass=posixAccount)', ['uidNumber'])

	# store all uidNumbers in list
	allUIDNumbers = []
	for result in results:
		allUIDNumbers.append(int(result[1]['uidNumber'][0]))

	rv = -1
	for x in range(int(minUIDNumber), int(maxUIDNumber)):
		if x not in allUIDNumbers:
			rv = x
			break

	return rv